import os
import multiprocessing
from pathlib import Path
import shutil

import h5py
from xrf_tomo import process_proj, make_single_hdf

# sys.stderr has been replaced by vtkPythonStdStreamCaptureHelper
# As such, it does not have an "isatty()" function. Therefore, turn off
# the tty check in the progress library.
try:
    import progress
    progress.Infinite.check_tty = False
except Exception:
    # Maybe a new version of pyxrf removed dependency on progress...
    # Or maybe progress refactored. Either way, let it go...
    pass


def ic_names(working_directory):
    # Find any HDF5 file in the working directory and grab the ic names
    files = [x for x in Path(working_directory).iterdir() if x.suffix == '.h5']
    if not files:
        raise Exception(f'No .h5 files found in "{working_directory}"')

    first_file = files[0]

    with h5py.File(first_file, 'r') as rf:
        names = rf["xrfmap"]["scalers"]["name"]
        names = [x.decode() for x in names]
        return names


def process_projections(working_directory, parameters_file_name, log_file_name,
                        ic_name, skip_processed, output_directory):
    # FIXME: this shouldn't be necessary. When we get time, figure out why
    # we have to specify these paths. Currently, if we do not, then the
    # multiprocessing will encounter errors about missing standard python
    # libraries.
    fix_python_paths()

    # Because we are running an embedded tomviz, multiprocessing can't
    # find the python executable. Therefore, we need to try to find it
    # automatically and tell multiprocessing where it is.
    python_exec = find_python_executable()
    multiprocessing.set_executable(python_exec)

    kwargs = {
        'wd': working_directory,
        'fn_param': parameters_file_name,
        'fn_log': log_file_name,
        'ic_name': ic_name,
        'skip_processed': skip_processed,
    }
    process_proj(**kwargs)

    kwargs = {
        'fn': 'tomo.h5',
        'fn_log': log_file_name,
        'wd_src': working_directory,
        'wd_dest': output_directory,
        'ic_name': ic_name,
    }
    make_single_hdf(**kwargs)

    # Copy the csv file into the output directory
    log_file_path = Path(log_file_name)
    if not log_file_path.is_absolute():
        log_file_path = Path(working_directory).resolve() / log_file_path

    output_file_path = Path(output_directory).resolve() / log_file_path.name
    if log_file_path != output_file_path:
        # Copy the csv file into the output directory
        shutil.copyfile(log_file_path, output_file_path)


def fix_python_paths():
    # FIXME: this shouldn't be necessary, but otherwise, we get errors
    # indicating that python standard libraries couldn't be found.
    # It seems the embedded python in our conda package doesn't set up
    # its sys.path correctly.
    python_path_var = os.environ.get('PYTHONPATH', '')
    python_paths = python_path_var.split(':') if python_path_var else []

    conda_path = Path(os.environ.get('CONDA_PREFIX', ''))
    need_to_add = [
        conda_path / 'lib/python3.7/site-packages',
        conda_path / 'lib/python3.7/lib-dynload',
    ]

    for path in need_to_add:
        if str(path) not in python_paths:
            python_paths.append(str(path))

    os.environ['PYTHONPATH'] = ':'.join(python_paths)


def find_python_executable():
    # Since we are in tomviz, sys.executable does not work
    failure_msg = (
        'Could not find python executable for multiprocessing. '
        'Please set it to the environment variable "TOMVIZ_PYTHON_EXECUTABLE" '
        'and restart tomviz.'
    )

    # First, check if the user set it via an environment variable
    path = os.environ.get('TOMVIZ_PYTHON_EXECUTABLE')
    if path and Path(path).exists():
        return path

    # Next, try to find it via the conda prefix
    conda_path = os.environ.get('CONDA_PREFIX')
    if conda_path is None:
        raise Exception(failure_msg)

    try_list = [
        Path(conda_path) / 'bin' / 'python',
        Path(conda_path) / 'bin' / 'python3',
        Path(conda_path) / 'bin' / 'python.exe',
    ]
    for path in try_list:
        if path.exists():
            return str(path)

    raise Exception(failure_msg)



# -*- coding: utf-8 -*-

###############################################################################
# This source file is part of the Tomviz project, https://tomviz.org/.
# It is released under the 3-Clause BSD License, see "LICENSE".
###############################################################################
import inspect
import pkgutil

import tomviz.io
import tomviz.io.formats


def list_python_readers():

    readers = []
    for importer, name, _ in pkgutil.iter_modules(tomviz.io.formats.__path__,
                                                  prefix='tomviz.io.formats.'):
        m = importer.find_spec(name).loader.load_module()
        for _, c in inspect.getmembers(m, inspect.isclass):
            if inspect.getmodule(c) is m:
                if issubclass(c, tomviz.io.Reader):
                    readers.append(
                        [
                            c.file_type().display_name,
                            c.file_type().extensions,
                            c
                        ]
                    )
    return readers


def create_reader_instance(reader_class):
    return reader_class()


def execute_reader(obj, path):
    return obj.read(path)


def list_python_writers():
    writers = []
    for importer, name, _ in pkgutil.iter_modules(tomviz.io.formats.__path__,
                                                  prefix='tomviz.io.formats.'):
        m = importer.find_spec(name).loader.load_module()
        for _, c in inspect.getmembers(m, inspect.isclass):
            if inspect.getmodule(c) is m:
                if issubclass(c, tomviz.io.Writer):
                    writers.append(
                        [
                            c.file_type().display_name,
                            c.file_type().extensions,
                            c
                        ]
                    )
    return writers


def create_writer_instance(writer_class):
    return writer_class()


def execute_writer(obj, path, data):
    obj.write(path, data)



# -*- coding: utf-8 -*-

###############################################################################
# This source file is part of the Tomviz project, https://tomviz.org/.
# It is released under the 3-Clause BSD License, see "LICENSE".
###############################################################################

from tomviz._internal import in_application
from tomviz._internal import require_internal_mode
from tomviz._internal import with_dataset
from tomviz._internal import with_vtk_dataobject

# Dictionary going from VTK array type to ITK type
_vtk_to_itk_types = None

# Dictionary mapping from VTK array numeric type to the VTK array numeric type
# supported by the ITK wrapping. Used to cast data sets to a supported type
# prior to converting them to ITK images.
_vtk_cast_types = None

# Dictionary mapping from ITK image type to Python numeric type.
# Used for casting Python values to a suitable type for certain filters.
_itkctype_to_python_types = None

# Map between VTK numeric type and Python numeric type
_vtk_to_python_types = None


def vtk_itk_type_map():
    """Set up mappings between VTK image types and available ITK image
    types."""
    global _vtk_to_itk_types

    if _vtk_to_itk_types is None:
        import itk
        _vtk_to_itk_types = {}

        type_map = {
            'vtkUnsignedCharArray': 'UC3',
            'vtkCharArray': 'SC3',
            'vtkUnsignedShortArray': 'US3',
            'vtkShortArray': 'SS3',
            'vtkUnsignedIntArray': 'UI3',
            'vtkIntArray': 'SI3',
            'vtkFloatArray': 'F3',
            'vtkDoubleArray': 'D3'
        }

        for (vtk_type, image_type) in type_map.items():
            try:
                _vtk_to_itk_types[vtk_type] = getattr(itk.Image, image_type)
            except AttributeError:
                pass

    return _vtk_to_itk_types


def vtk_cast_map():
    """Set up mapping between VTK array numeric types to VTK numeric types
    that correspond to supported ITK numeric ctypes supported by the ITK
    wrapping."""
    require_internal_mode()

    global _vtk_cast_types

    if _vtk_cast_types is None:
        import itk
        _vtk_cast_types = {}

        # Map from VTK array type to list of possible types to which they can be
        # converted, listed in order of preference based on representability
        # and memory size required.
        from vtkmodules.util import vtkConstants
        type_map = {
            vtkConstants.VTK_UNSIGNED_CHAR: [
                'unsigned char',
                'unsigned short',
                'unsigned int',
                'unsigned long',
                'signed short',
                'signed int',
                'signed long',
                'float',
                'double'
            ],
            vtkConstants.VTK_CHAR: [
                'signed char',
                'signed short',
                'signed int',
                'signed long',
                'float',
                'double'
            ],
            vtkConstants.VTK_SIGNED_CHAR: [
                'signed char',
                'signed short',
                'signed int',
                'signed long',
                'float',
                'double'
            ],
            vtkConstants.VTK_UNSIGNED_SHORT: [
                'unsigned short',
                'unsigned int',
                'unsigned long',
                'signed int',
                'signed long',
                'float',
                'double'
            ],
            vtkConstants.VTK_SHORT: [
                'signed short',
                'signed int',
                'signed long',
                'float',
                'double'
            ],
            vtkConstants.VTK_UNSIGNED_INT: [
                'unsigned int',
                'unsigned long',
                'signed long',
                'float',
                'double'
            ],
            vtkConstants.VTK_INT: [
                'signed int',
                'signed long',
                'float',
                'double'
            ],
            vtkConstants.VTK_FLOAT: [
                'float',
                'double'
            ],
            vtkConstants.VTK_DOUBLE: [
                'double',
                'float'
            ]
        }

        # Map ITK ctype back to VTK type
        ctype_to_vtk = {
            'unsigned char': vtkConstants.VTK_UNSIGNED_CHAR,
            'signed char': vtkConstants.VTK_CHAR,
            'unsigned short': vtkConstants.VTK_UNSIGNED_SHORT,
            'signed short': vtkConstants.VTK_SHORT,
            'unsigned int': vtkConstants.VTK_UNSIGNED_INT,
            'signed int': vtkConstants.VTK_INT,
            'unsigned long': vtkConstants.VTK_UNSIGNED_LONG,
            'signed long': vtkConstants.VTK_LONG,
            'float': vtkConstants.VTK_FLOAT,
            'double': vtkConstants.VTK_DOUBLE
        }

        # Import build options from ITK. Explicitly reference itk.Vector so
        # that the itk::Vector type information is available when
        # itk.BuildOptions is imported, otherwise we get an exception when
        # importing itkBuildOptions. This is a workaround for a bug in ITK.
        itk.Vector
        import itkBuildOptions

        # Select the best supported type available in the wrapping.
        for (vtk_type, possible_image_types) in type_map.items():
            type_map[vtk_type] = None
            for possible_type in possible_image_types:
                if itk.ctype(possible_type) in itkBuildOptions.SCALARS:
                    _vtk_cast_types[vtk_type] = ctype_to_vtk[possible_type]
                    break

    return _vtk_cast_types


def get_python_voxel_type(dataset):
    """Return the Python type that can represent the voxel type in the dataset.
    The dataset can be either a VTK dataset or an ITK image.
    """

    if in_application():
        # Try treating dataset as a VTK data set first.
        try:
            # Set up map between VTK data type and Python type
            global _vtk_to_python_types

            if _vtk_to_python_types is None:
                from vtkmodules.util import vtkConstants
                _vtk_to_python_types = {
                    vtkConstants.VTK_UNSIGNED_CHAR: int,
                    vtkConstants.VTK_CHAR: int,
                    vtkConstants.VTK_UNSIGNED_SHORT: int,
                    vtkConstants.VTK_SHORT: int,
                    vtkConstants.VTK_UNSIGNED_INT: int,
                    vtkConstants.VTK_INT: int,
                    vtkConstants.VTK_UNSIGNED_LONG: int,
                    vtkConstants.VTK_LONG: int,
                    vtkConstants.VTK_FLOAT: float,
                    vtkConstants.VTK_DOUBLE: float
                }

            pd = dataset.GetPointData()
            scalars = pd.GetScalars()

            return _vtk_to_python_types[scalars.GetDataType()]
        except AttributeError:
            pass

    # If the above fails, treat dataset as an ITK image.
    try:
        # Set up map between ITK ctype and Python type.
        global _itkctype_to_python_types

        if _itkctype_to_python_types is None:
            import itkTypes

            _itkctype_to_python_types = {
                itkTypes.F: float,
                itkTypes.D: float,
                itkTypes.LD: float,
                itkTypes.UC: int,
                itkTypes.US: int,
                itkTypes.UI: int,
                itkTypes.UL: int,
                itkTypes.SC: int,
                itkTypes.SS: int,
                itkTypes.SI: int,
                itkTypes.SL: int,
                itkTypes.B: int
            }

        import itkExtras

        # Incantation for obtaining voxel type in ITK image
        ctype = itkExtras.template(type(dataset))[1][0]
        return _itkctype_to_python_types[ctype]
    except AttributeError as attribute_error:
        print("Could not get Python voxel type for dataset %s"
              % type(dataset))
        print(attribute_error)


@with_vtk_dataobject
def convert_vtk_to_itk_image(vtk_image_data, itk_pixel_type=None):
    """Get an ITK image from the provided vtkImageData object.
    This image can be passed to ITK filters."""

    # Save the VTKGlue optimization for later
    #------------------------------------------
    #itk_import = itk.VTKImageToImageFilter[image_type].New()
    #itk_import.SetInput(vtk_image_data)
    #itk_import.Update()
    #itk_image = itk_import.GetOutput()
    #itk_image.DisconnectPipeline()
    #------------------------------------------
    import itk
    import itkTypes
    from vtkmodules.util import vtkConstants
    from vtkmodules.vtkImagingCore import vtkImageCast
    from tomviz import utils

    itk_to_vtk_type_map = {
        itkTypes.F: vtkConstants.VTK_FLOAT,
        itkTypes.D: vtkConstants.VTK_DOUBLE,
        itkTypes.LD: vtkConstants.VTK_DOUBLE,
        itkTypes.UC: vtkConstants.VTK_UNSIGNED_CHAR,
        itkTypes.US: vtkConstants.VTK_UNSIGNED_SHORT,
        itkTypes.UI: vtkConstants.VTK_UNSIGNED_INT,
        itkTypes.UL: vtkConstants.VTK_UNSIGNED_LONG,
        itkTypes.SC: vtkConstants.VTK_CHAR,
        itkTypes.SS: vtkConstants.VTK_SHORT,
        itkTypes.SI: vtkConstants.VTK_INT,
        itkTypes.SL: vtkConstants.VTK_LONG,
        itkTypes.B: vtkConstants.VTK_INT
    }

    # See if we need to cast to a wrapped type in ITK.
    src_type = vtk_image_data.GetScalarType()

    if itk_pixel_type is None:
        dst_type = vtk_cast_map()[src_type]
    else:
        dst_type = vtk_cast_map()[itk_to_vtk_type_map[itk_pixel_type]]
    if src_type != dst_type:
        caster = vtkImageCast()
        caster.SetOutputScalarType(dst_type)
        caster.ClampOverflowOn()
        caster.SetInputData(vtk_image_data)
        caster.Update()
        vtk_image_data = caster.GetOutput()

    array = utils.get_array(vtk_image_data, order='C')

    image_type = _get_itk_image_type(vtk_image_data)
    itk_converter = itk.PyBuffer[image_type]
    itk_image = itk_converter.GetImageFromArray(array)
    spacing = vtk_image_data.GetSpacing()
    origin = vtk_image_data.GetOrigin()
    itk_image.SetSpacing(spacing)
    itk_image.SetOrigin(origin)

    # Persist a reference to the source vtk_image_data, which is necessary since
    # VTK and ITK are using Python Buffer-Protocol NumPy array views
    itk_image.vtk_image_data = vtk_image_data

    return itk_image


@with_vtk_dataobject
def set_array_from_itk_image(dataobject, itk_image):
    """Set dataobject array from an ITK image."""
    itk_output_image_type = type(itk_image)

    # Save the VTKGlue optimization for later
    #------------------------------------------
    # Export the ITK image to a VTK image. No copying should take place.
    #export_filter = itk.ImageToVTKImageFilter[itk_output_image_type].New()
    #export_filter.SetInput(itk_image_data)
    #export_filter.Update()

    # Get scalars from the temporary image and copy them to the data set
    #result_image = export_filter.GetOutput()
    #filter_array = result_image.GetPointData().GetArray(0)

    # Make a new instance of the array that will stick around after this
    # filters in this script are garbage collected
    #new_array = filter_array.NewInstance()
    #new_array.DeepCopy(filter_array) # Should be able to shallow copy?
    #new_array.SetName(name)
    #------------------------------------------
    import itk
    from . import utils
    result = itk.PyBuffer[
        itk_output_image_type].GetArrayFromImage(itk_image)
    result = result.copy()
    utils.set_array(dataobject, result, isFortran=False)


@with_dataset
def get_label_object_attributes(dataset, progress_callback=None):
    """Compute shape attributes of integer-labeled objects in a dataset. Returns
    an ITK shape label map. An optional progress_callback function can be passed
    in. This callback is expected to take one argument, a floating-point number
    in the range [0, 1] that represents the progress amount. It returns a value
    indicating whether the caller should be cancelled.
    """

    try:
        import itk

        # Get an ITK image from the data set
        itk_image = dataset_to_itk_image(dataset)
        itk_image_type = type(itk_image)

        # Get an appropriate LabelImageToShapelLabelMapFilter type for the
        # input.
        inputTypes = [x[0] for x in list(itk.LabelImageToShapeLabelMapFilter.keys())] # noqa
        filterTypeIndex = inputTypes.index(itk_image_type)
        if filterTypeIndex < 0:
            raise Exception("No suitable filter type for input type %s" %
                            type(itk_image_type))

        # Now take the connected components results and compute things like
        # volume and surface area.
        shape_filter = \
            list(itk.LabelImageToShapeLabelMapFilter.values())[filterTypeIndex].New() # noqa
        shape_filter.SetInput(itk_image)

        def progress_func():
            progress = shape_filter.GetProgress()
            if progress_callback is not None:
                abort = progress_callback(progress)
                if abort:
                    shape_filter.AbortGenerateDataOn()

        progress_observer = itk.PyCommand.New()
        progress_observer.SetCommandCallable(progress_func)
        shape_filter.AddObserver(itk.ProgressEvent(), progress_observer)

        try:
            shape_filter.Update()
        except RuntimeError:
            return None

        label_map = shape_filter.GetOutput()
        return label_map
    except Exception as exc:
        print("Problem encountered while running label_object_attributes")
        raise(exc)


@with_vtk_dataobject
def _get_itk_image_type(vtk_image_data):
    """
    Get an ITK image type corresponding to the provided vtkImageData object.
    """
    image_type = None

    # Get the scalars
    pd = vtk_image_data.GetPointData()
    scalars = pd.GetScalars()

    vtk_class_name = scalars.GetClassName()

    try:
        image_type = vtk_itk_type_map()[vtk_class_name]
    except KeyError:
        raise Exception('No ITK type known for %s' % vtk_class_name)

    return image_type


def observe_filter_progress(transform, filter, start_pct, end_pct):
    assert start_pct < end_pct
    pct_diff = end_pct - start_pct

    def progress_func():
        progress = filter.GetProgress()
        transform.progress.value = start_pct + int(progress * pct_diff)
        if transform.canceled:
            filter.AbortGenerateDataOn()

    import itk
    progress_observer = itk.PyCommand.New()
    progress_observer.SetCommandCallable(progress_func)
    filter.AddObserver(itk.ProgressEvent(), progress_observer)


@with_dataset
def dataset_to_itk_image(dataset):

    import itk

    itk_image = itk.GetImageViewFromArray(dataset.active_scalars)

    if dataset.spacing is not None:
        itk_image.SetSpacing(dataset.spacing)

    return itk_image


def set_itk_image_on_dataset(itk_image, dataset, dtype=None):
    # Write the itk image data to the dataset

    import itk

    if dtype is None:
        array = itk.GetArrayFromImage(itk_image)
    else:
        array = itk.PyBuffer[dtype].GetArrayFromImage(itk_image)

    # Transpose the data to Fortran indexing
    dataset.active_scalars = array.transpose([2, 1, 0])



def transform(dataset):
    """
    For each tilt image, the method calculates its histogram
    and then chooses the highest peak as the background level and subtracts it
    from the image.
    It does NOT set negative pixels to zero.
    """

    import numpy as np

    data = dataset.active_scalars # Get data as numpy array.

    if data is None: #Check if data exists
        raise RuntimeError("No data array found!")

    data_bs = data.astype(np.float32) # Change tilt series type to float.

    for i in range(0, data.shape[2]):
        (hist, bins) = np.histogram(data[:, :, i].flatten(), 256)
        data_bs[:, :, i] = data_bs[:, :, i] - bins[np.argmax(hist)]

    dataset.active_scalars = data_bs



# -*- coding: utf-8 -*-

###############################################################################
# This source file is part of the Tomviz project, https://tomviz.org/.
# It is released under the 3-Clause BSD License, see "LICENSE".
###############################################################################
from abc import ABCMeta, abstractmethod


class FileType(object):
    """
    Container class for file type information.
    """
    def __init__(self, display_name=None, extensions=None):
        self.display_name = display_name
        self.extensions = extensions

    def __str__(self):
        return "%s (%s)" % (self.display_name,
                            " ".join(["*."+ext for ext in self.extensions]))


class IOBase(object, metaclass=ABCMeta):

    @staticmethod
    @abstractmethod
    def file_type():
        """
        :returns An instance of the FileFormat class. This is used to associate
        a file type with a reader.
        :rtype tomviz.io.FileType
        """


class Reader(IOBase, metaclass=ABCMeta):
    """
    The base reader class from which readers should be derived.
    """

    """
    Set to True if reader supports loading image stacks.
    """
    supports_stacks = False

    @abstractmethod
    def read(self, file):
        """
        :returns Return a vtkDataObject containing the scalars read.
        :param file: the path or file object to read from
        :rtype vtk.vtkDataObject
        """


class Writer(IOBase, metaclass=ABCMeta):
    """
    The base reader class from which writers should be derived.
    """

    @abstractmethod
    def write(self, file, data):
        """
        :param file: the path or file object to write to
        :param data: The data to write to the file.
        :type data: The vtkDataObject instance.
        """



# -*- coding: utf-8 -*-

###############################################################################
# This source file is part of the Tomviz project, https://tomviz.org/.
# It is released under the 3-Clause BSD License, see "LICENSE".
###############################################################################
import numpy as np
import scipy.io

from tomviz.io import FileType, IOBase, Reader

import tomviz.utils

from vtk import vtkImageData


class MatlabBase(IOBase):

    @staticmethod
    def file_type():
        return FileType('MATLAB binary format', ['mat'])


class MatlabReader(Reader, MatlabBase):

    def read(self, path):
        try:
            mat_dict = scipy.io.loadmat(path)
        except NotImplementedError as e:
            # matlab v7.3 requires h5py to load
            if 'matlab v7.3' in str(e).lower():
                print('Tomviz does not currently support matlab v7.3 files')
                print('Please convert the file to matlab v7.2 or earlier')
                return vtkImageData()
            raise

        data = None
        for item in mat_dict.values():
            # Assume only one 3D array per file
            if isinstance(item, np.ndarray):
                if len(item.shape) == 3:
                    data = item
                    break

        if data is None:
            return vtkImageData()

        image_data = vtkImageData()
        (x, y, z) = data.shape
        image_data.SetOrigin(0, 0, 0)
        image_data.SetSpacing(1, 1, 1)
        image_data.SetExtent(0, x - 1, 0, y - 1, 0, z - 1)
        tomviz.utils.set_array(image_data, data)

        return image_data



def transform(dataset):
    """Downsample volume by a factor of 2"""

    from tomviz import utils
    import scipy.ndimage
    import numpy as np

    array = dataset.active_scalars

    # Downsample the dataset x2 using order 1 spline (linear)
    # Calculate out array shape
    zoom = (0.5, 0.5, 0.5)
    result_shape = utils.zoom_shape(array, zoom)
    result = np.empty(result_shape, array.dtype, order='F')
    scipy.ndimage.interpolation.zoom(array, zoom,
                                     output=result, order=1,
                                     mode='constant', cval=0.0,
                                     prefilter=False)

    # Set the result as the new scalars.
    dataset.active_scalars = result

    # Update tilt angles if dataset is a tilt series.
    try:
        tilt_angles = dataset.tilt_angles
        result_shape = utils.zoom_shape(tilt_angles, 0.5)
        result = np.empty(result_shape, array.dtype, order='F')
        tilt_angles = scipy.ndimage.interpolation.zoom(tilt_angles, 0.5,
                                                       output=result)
        dataset.tilt_angles = result
    except: # noqa
        # TODO What exception are we ignoring?
        pass



from vtkmodules.util.vtkAlgorithm import VTKPythonAlgorithmBase

from .. import print_error

try:
    import numpy as np
    import openpmd_api as io
    _has_openpmd = True
except ImportError as ie:
    print_error(
        "Missing required Python modules/packages. Algorithms in this module may "
        "not work as expected! \n {0}".format(ie))
    _has_openpmd = False

def createModifiedCallback(anobject):
    import weakref
    weakref_obj = weakref.ref(anobject)
    anobject = None
    def _markmodified(*args, **kwars):
        o = weakref_obj()
        if o is not None:
            o.Modified()
    return _markmodified

class openPMDReader(VTKPythonAlgorithmBase):
    """A reader that reads openPMD format."""
    def __init__(self):
        VTKPythonAlgorithmBase.__init__(self, nInputPorts=0, nOutputPorts=2, outputType='vtkPartitionedDataSet')
        self._filename = None
        self._timevalues = None
        self._series = None
        self._timemap = {}

        from vtkmodules.vtkCommonCore import vtkDataArraySelection
        self._arrayselection = vtkDataArraySelection()
        self._arrayselection.AddObserver("ModifiedEvent", createModifiedCallback(self))

        self._speciesselection = vtkDataArraySelection()
        self._speciesselection.AddObserver("ModifiedEvent", createModifiedCallback(self))

        self._particlearrayselection = vtkDataArraySelection()
        self._particlearrayselection.AddObserver("ModifiedEvent", createModifiedCallback(self))

    def _get_update_time(self, outInfo):
        from vtkmodules.vtkCommonExecutionModel import vtkStreamingDemandDrivenPipeline
        executive = vtkStreamingDemandDrivenPipeline
        timevalues = self._timevalues
        if timevalues is None or len(timevalues) == 0:
            return None
        elif outInfo.Has(executive.UPDATE_TIME_STEP()) and len(timevalues) > 0:
            utime = outInfo.Get(executive.UPDATE_TIME_STEP())
            dtime = timevalues[0]
            for atime in timevalues:
                if atime > utime:
                    return dtime
                else:
                    dtime = atime
            return dtime
        else:
            assert(len(timevalues) > 0)
            return timevalues[0]

    def _get_array_selection(self):
        return self._arrayselection

    def _get_particle_array_selection(self):
        return self._particlearrayselection

    def _get_species_selection(self):
        return self._speciesselection

    def SetFileName(self, name):
        """Specify filename for the file to read."""
        if self._filename != name:
            self._filename = name
            self._timevalues = None
            if self._series:
                self._series = None
            self.Modified()

    def GetTimestepValues(self):
        return self._timevalues()

    def GetDataArraySelection(self):
        return self._get_array_selection()

    def GetSpeciesSelection(self):
        return self._get_species_selection()

    def GetParticleArraySelection(self):
        return self._get_particle_array_selection()

    def FillOutputPortInformation(self, port, info):
        from vtkmodules.vtkCommonDataModel import vtkDataObject
        if port == 0:
            info.Set(vtkDataObject.DATA_TYPE_NAME(), "vtkPartitionedDataSet")
        else:
            info.Set(vtkDataObject.DATA_TYPE_NAME(), "vtkPartitionedDataSetCollection")
        return 1

    def RequestInformation(self, request, inInfoVec, outInfoVec):
        global _has_openpmd
        if not _has_openpmd:
            print_error("Required Python module 'openpmd_api' missing!")
            return 0

        from vtkmodules.vtkCommonExecutionModel import vtkStreamingDemandDrivenPipeline, vtkAlgorithm
        executive = vtkStreamingDemandDrivenPipeline
        for i in (0,1):
            outInfo = outInfoVec.GetInformationObject(i)
            outInfo.Remove(executive.TIME_STEPS())
            outInfo.Remove(executive.TIME_RANGE())
            outInfo.Set(vtkAlgorithm.CAN_HANDLE_PIECE_REQUEST(), 1)

        # Why is this a string when it is None?
        if self._filename == 'None':
            return 1

        mfile = open(self._filename, "r")
        pattern = mfile.readlines()[0][0:-1]
        del mfile


        import os
        if not self._series:
            self._series = io.Series(os.path.dirname(self._filename)+'/'+pattern, io.Access_Type.read_only)
        # This is how we get time values and arrays
        self._timemap = {}
        timevalues = []
        arrays = set()
        particles = set()
        species = set()
        for idx, iteration in self._series.iterations.items():
            time = iteration.time()
            timevalues.append(time)
            self._timemap[time] = idx
            arrays.update([
                mesh_name
                for mesh_name, mesh in iteration.meshes.items()])
            particles.update([
                species_name + "_" + record_name
                for species_name, species in iteration.particles.items()
                for record_name, record in species.items()
                ])
            species.update([
                species_name
                for species_name, _ in iteration.particles.items()
                ])

        for array in arrays:
            self._arrayselection.AddArray(array)
        for particle_array in particles:
            self._particlearrayselection.AddArray(particle_array)
        for species_name in species:
            self._speciesselection.AddArray(species_name)

        timesteps = list(self._series.iterations)
        self._timevalues = timevalues
        if len(timevalues) > 0:
            for i in (0,1):
                outInfo = outInfoVec.GetInformationObject(i)
                for t in timevalues:
                    outInfo.Append(executive.TIME_STEPS(), t)
                outInfo.Append(executive.TIME_RANGE(), timevalues[0])
                outInfo.Append(executive.TIME_RANGE(), timevalues[-1])
        return 1

    def _get_array_and_component(self, itr, name):
        for mesh_name, mesh in itr.meshes.items():
            if mesh_name == name:
                return (mesh_name, None)
            for comp_name, _ in mesh.items():
                if name == mesh_name + "_" + comp_name:
                    return (mesh_name, comp_name)
        return (None, None)

    def _get_particle_array_and_component(self, itr, name):
        for species_name, species in itr.particles.items():
            for record_name, record in species.items():
                if name == species_name + "_" + record_name:
                    return (species_name, record_name)
        return (None, None)

    def _load_array(self, var, chunk_offset, chunk_extent):
        arrays = []
        for name, scalar in var.items():
            comp = scalar.load_chunk(chunk_offset, chunk_extent)
            self._series.flush()
            comp = comp * scalar.unit_SI
            arrays.append(comp)

        ncomp = len(var)
        if ncomp > 1:
            flt = np.ravel(arrays, order='F')
            return flt.reshape((flt.shape[0]//ncomp, ncomp))
        else:
            return arrays[0].flatten(order='F')

    def _find_array(self, itr, name):
        var = itr.meshes[name]
        theta_modes = None
        if var.geometry == io.Geometry.thetaMode:
            theta_modes = 3 # hardcoded, parse from geometry_parameters
        return (var, np.array(var.grid_spacing)*var.grid_unit_SI, np.array(var.grid_global_offset)*var.grid_unit_SI, theta_modes)

    def _get_num_particles(self, itr, species):
        sp = itr.particles[species]
        var = sp["position"]
        return var['x'].shape[0]

    def _load_particle_array(self, itr, species, name, start, end):
        sp = itr.particles[species]
        var = sp[name]
        arrays = []
        for name, scalar in var.items():
            comp = scalar.load_chunk([start], [end-start+1])
            self._series.flush()
            comp = comp * scalar.unit_SI
            arrays.append(comp)

        ncomp = len(var)
        if ncomp > 1:
            flt = np.ravel(arrays, order='F')
            return flt.reshape((flt.shape[0]//ncomp, ncomp))
        else:
            return arrays[0]

    def _load_particles(self, itr, species, start, end):
        sp = itr.particles[species]
        var = sp["position"]
        ovar = sp["positionOffset"]
        position_arrays = []
        for name, scalar in var.items():
            pos = scalar.load_chunk([start], [end-start+1])
            self._series.flush()
            pos = pos * scalar.unit_SI
            off = ovar[name].load_chunk([start], [end-start+1])
            self._series.flush()
            off = off * ovar[name].unit_SI
            position_arrays.append(pos + off)

        flt = np.ravel(position_arrays, order='F')
        return flt.reshape((flt.shape[0]//3, 3))

    def _load_species(self, itr, species, arrays, piece, npieces, ugrid):
        nparticles = self._get_num_particles(itr, species)
        nlocalparticles = nparticles // npieces
        start = nlocalparticles * piece
        end = start + nlocalparticles - 1
        if piece == npieces - 1:
            end = nparticles - 1
        pts = self._load_particles(itr, species, start, end)
        npts = pts.shape[0]
        ugrid.Points = pts
        for array in arrays:
            if array[1] == 'position' or array[1] == 'positionOffset':
                continue
            ugrid.PointData.append(
                self._load_particle_array(itr, array[0], array[1], start, end),
                array[1])
        from vtkmodules.vtkCommonDataModel import vtkCellArray
        ca = vtkCellArray()
        if npts < np.iinfo(np.int32).max:
            dtype = np.int32
        else:
            dtype = np.int64
        offsets = np.linspace(0, npts, npts+1, dtype=dtype)
        cells = np.linspace(0, npts-1, npts, dtype=dtype)
        from vtkmodules.numpy_interface import dataset_adapter
        offsets = dataset_adapter.numpyTovtkDataArray(offsets)
        offsets2 = offsets.NewInstance()
        offsets2.DeepCopy(offsets)
        cells = dataset_adapter.numpyTovtkDataArray(cells)
        cells2 = cells.NewInstance()
        cells2.DeepCopy(cells)
        ca.SetData(offsets2, cells2)
        from vtkmodules.util import vtkConstants
        ugrid.VTKObject.SetCells(vtkConstants.VTK_VERTEX, ca)


    def RequestData(self, request, inInfoVec, outInfoVec):
        global _has_openpmd
        if not _has_openpmd:
            print_error("Required Python module 'openpmd_api' missing!")
            return 0

        from vtkmodules.vtkCommonDataModel import vtkImageData, vtkUnstructuredGrid
        from vtkmodules.vtkCommonDataModel import vtkPartitionedDataSet,vtkPartitionedDataSetCollection
        from vtkmodules.vtkCommonExecutionModel import vtkExtentTranslator, vtkStreamingDemandDrivenPipeline
        from vtkmodules.numpy_interface import dataset_adapter as dsa

        executive = vtkStreamingDemandDrivenPipeline
        output = vtkPartitionedDataSet.GetData(outInfoVec, 0)
        poutput = vtkPartitionedDataSetCollection.GetData(outInfoVec, 1)
        outInfo = outInfoVec.GetInformationObject(0)
        piece = outInfo.Get(executive.UPDATE_PIECE_NUMBER())
        npieces = outInfo.Get(executive.UPDATE_NUMBER_OF_PIECES())
        nghosts = outInfo.Get(executive.UPDATE_NUMBER_OF_GHOST_LEVELS())
        et = vtkExtentTranslator()

        data_time = self._get_update_time(outInfo)
        idx = self._timemap[data_time]

        itr = self._series.iterations[idx]
        arrays = []
        narrays = self._arrayselection.GetNumberOfArrays()
        for i in range(narrays):
            if self._arrayselection.GetArraySetting(i):
                name = self._arrayselection.GetArrayName(i)
                arrays.append((name, self._find_array(itr, name)))
        shp = None
        spacing = None
        theta_modes = None
        grid_offset = None
        for _, ary in arrays:
            var = ary[0]
            for name, scalar in var.items():
                shape = scalar.shape
                break
            spc = list(ary[1])
            if not spacing:
                spacing = spc
            elif spacing != spc: # all meshes need to have the same spacing
                return 0
            offset = list(ary[2])
            if not grid_offset:
                grid_offset = offset
            elif grid_offset != offset: # all meshes need to have the same spacing
                return 0
            if not shp:
                shp = shape
            elif shape != shp: # all arrays needs to have the same shape
                return 0
            if not theta_modes:
                theta_modes = ary[3]

        if theta_modes:
            et.SetWholeExtent(0, shp[0]-1, 0, shp[1]-1, 0, shp[2]-1)
            et.SetSplitModeToZSlab() # note: Y and Z are both fine
            et.SetPiece(piece)
            et.SetNumberOfPieces(npieces)
            # et.SetGhostLevel(nghosts)
            et.PieceToExtentByPoints()
            ext = et.GetExtent()

            chunk_offset = [ext[0], ext[2], ext[4]]
            chunk_extent = [ext[1] - ext[0] + 1, ext[3] - ext[2] + 1, ext[5] - ext[4] + 1]

            data = []
            nthetas = 100  # user parameter
            thetas = np.linspace(0., 2.*np.pi, nthetas)
            chunk_cyl_shape = (chunk_extent[1], chunk_extent[2], nthetas)  # z, r, theta
            for name, var in arrays:
                cyl_values = np.zeros(chunk_cyl_shape)
                values = self._load_array(var[0], chunk_offset, chunk_extent)
                self._series.flush()

                print(chunk_cyl_shape)
                print(values.shape)
                print("+++++++++++")
                for ntheta in range(nthetas):
                    cyl_values[:, :, ntheta] += values[0, :, :]
                data.append((name, cyl_values))
                # add all other modes via loop
                # for m in range(theta_modes):

            cyl_spacing = [spacing[0], spacing[1], thetas[1]-thetas[0]]

            z_coord = np.linspace(0., cyl_spacing[0] * chunk_cyl_shape[0], chunk_cyl_shape[0])
            r_coord = np.linspace(0., cyl_spacing[1] * chunk_cyl_shape[1], chunk_cyl_shape[1])
            t_coord = thetas

            # to cartesian
            print(z_coord.shape, r_coord.shape, t_coord.shape)
            cyl_coords = np.meshgrid(r_coord, z_coord, t_coord)
            rs = cyl_coords[1]
            zs = cyl_coords[0]
            thetas = cyl_coords[2]

            y_coord = rs * np.sin(thetas)
            x_coord = rs * np.cos(thetas)
            z_coord = zs
            # mesh_pts = np.zeros((chunk_cyl_shape[0], chunk_cyl_shape[1], chunk_cyl_shape[2], 3))
            # mesh_pts[:, :, :, 0] = z_coord


            img = vtkImageData()
            img.SetExtent(
                chunk_offset[1], chunk_offset[1] + chunk_cyl_shape[0] - 1,
                chunk_offset[2], chunk_offset[2] + chunk_cyl_shape[1] - 1,
                0, nthetas-1)
            img.SetSpacing(cyl_spacing)

            imgw = dsa.WrapDataObject(img)
            output.SetPartition(0, img)
            for name, array in data:
                # print(array.shape)
                # print(array.transpose(2,1,0).flatten(order='C').shape)
                imgw.PointData.append(array.transpose(2,1,0).flatten(order='C'), name)


            # data = []
            # for name, var in arrays:
            #     unit_SI = var[0].unit_SI
            #     data.append((name, unit_SI * var[0].load_chunk(chunk_offset, chunk_extent)))
            # self._series.flush()

        else:
            et.SetWholeExtent(0, shp[0]-1, 0, shp[1]-1, 0, shp[2]-1)
            et.SetPiece(piece)
            et.SetNumberOfPieces(npieces)
            et.SetGhostLevel(nghosts)
            et.PieceToExtent()
            ext = et.GetExtent()

            chunk_offset = [ext[0], ext[2], ext[4]]
            chunk_extent = [ext[1] - ext[0] + 1, ext[3] - ext[2] + 1, ext[5] - ext[4] + 1]

            data = []
            for name, var in arrays:
                values = self._load_array(var[0], chunk_offset, chunk_extent)
                self._series.flush()
                data.append((name, values))

            img = vtkImageData()
            img.SetExtent(ext[0], ext[1], ext[2], ext[3], ext[4], ext[5])
            img.SetSpacing(spacing)
            img.SetOrigin(grid_offset)

            et.SetGhostLevel(0)
            et.PieceToExtent()
            ext = et.GetExtent()
            ext = [ext[0], ext[1], ext[2], ext[3], ext[4], ext[5]]
            img.GenerateGhostArray(ext)
            imgw = dsa.WrapDataObject(img)
            output.SetPartition(0, img)
            for name, array in data:
                imgw.PointData.append(array, name)

        itr = self._series.iterations[idx]
        array_by_species = {}
        narrays = self._particlearrayselection.GetNumberOfArrays()
        for i in range(narrays):
            if self._particlearrayselection.GetArraySetting(i):
                name = self._particlearrayselection.GetArrayName(i)
                names = self._get_particle_array_and_component(
                    itr, name)
                if names[0] and self._speciesselection.ArrayIsEnabled(names[0]):
                    if not names[0] in array_by_species:
                        array_by_species[names[0]] = []
                    array_by_species[names[0]].append(names)
        ids = 0
        for species, arrays in array_by_species.items():
            pds = vtkPartitionedDataSet()
            ugrid = vtkUnstructuredGrid()
            pds.SetPartition(0, ugrid)
            poutput.SetPartitionedDataSet(ids, pds)
            ids += 1
            self._load_species(
                itr, species, arrays, piece, npieces, dsa.WrapDataObject(ugrid))

        return 1



import json
import jsonpatch

from tomviz import operators, modules

from ._pipeline import PipelineStateManager

from ._utils import to_namespaces

op_class_attrs = ['description', 'label', 'script', 'type']


class InvalidStateError(RuntimeError):
    pass


class Base(object):
    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)


class Mortal(object):
    _dead = False

    def _kill(self):
        self._dead = True

    def __getattribute__(self, item):
        # Make an exception for the attribute we need to check or we will go
        # recursive!
        if item == '_dead':
            return super(Mortal, self).__getattribute__(item)

        if self._dead:
            raise InvalidStateError(
                "'%s' no longer represents a valid pipeline element." % self)
        else:
            return super(Mortal, self).__getattribute__(item)


class OperatorMeta(type):
    def __init__(cls, name, bases, dict):
        attrs = {}

        for k, v in dict.items():
            if k not in op_class_attrs:
                attrs[k] = property(lambda self: getattr(self, k),
                                    lambda self, value: setattr(self, k, value))
            else:
                attrs[k] = v

        super(OperatorMeta, cls).__init__(name, bases, attrs)


class ModuleMeta(type):
    def __new__(meta, name, parents, dct):
        attrs = {}

        attrs = {
            '_props': dct
        }

        return super(ModuleMeta, meta).__new__(meta, name, parents, attrs)


class Pipeline(Mortal):
    def __init__(self, datasource):
        self._datasource = datasource

    @property
    def dataSource(self):
        return self._datasource

    def _ds_path(self):
        from . import _pipeline_index
        pipeline_index = _pipeline_index(self._datasource)
        if pipeline_index < 0:
            raise Exception('Pipeline is not valid.')

        return '/dataSources/%d' % pipeline_index

    def pause(self):
        PipelineStateManager().pause_pipeline(self._ds_path())

    def resume(self):
        PipelineStateManager().resume_pipeline(self._ds_path())

    def execute(self):
        paused = self.paused()
        if paused:
            self.resume()

        PipelineStateManager().execute_pipeline(self._ds_path())

        if paused:
            self.pause()

    def paused(self):
        return PipelineStateManager().pipeline_paused(self._ds_path())


class Reader(Base, Mortal):
    pass


class DataSource(Base, Mortal):
    def __init__(self, **kwargs):
        self.reader = Reader(
            name=kwargs.pop('name', None),
            fileNames=kwargs.pop('fileNames', []),
            subsampleSettings=kwargs.pop('subsampleSettings', {}))

        self.operators = []
        self.modules = [modules.Outline(), modules.Slice()]
        super(DataSource, self).__init__(**kwargs)


class Operator(Base, Mortal):
    pass


class Module(Mortal):
    def __init__(self, **kwargs):
        args = self._props
        args.update(kwargs)

        for k, v in args.items():
            if isinstance(v, dict):
                v = to_namespaces(v)
            setattr(self, k, v)

    def _updates(self):
        from ._schemata import dump_module
        current_state = dump_module(self)
        patch = jsonpatch.JsonPatch.from_diff(self._props, current_state)
        patch = json.loads(patch.to_string())

        return patch


class Tomviz(object):
    def __init__(self, pipelines=None):
        self.pipelines = pipelines if pipelines else []


def module_json_to_classes(module_json):
    for name, info in module_json.items():
        info['type'] = name
        del info['id']
        # Default visibility to true
        info['properties']['visibility'] = True
        if not hasattr(modules, name):
            cls = ModuleMeta(name, (Module,), info)
            setattr(modules, name, cls)


def operator_json_to_classes(operator_json):
    for name, info in operator_json.items():
        del info['id']
        if not hasattr(operators, name):
            cls = OperatorMeta(name, (Operator,), info)
            setattr(operators, name, cls)


def init_modules():
    module_json_str = PipelineStateManager().module_json()
    module_json = json.loads(module_json_str)
    module_json_to_classes(module_json)


def init_operators():
    operator_json_str = PipelineStateManager().operator_json()
    operator_json = json.loads(operator_json_str)
    operator_json_to_classes(operator_json)



from enum import Enum
from paraview.simple import (
    Render,
    SaveScreenshot
)


class Palette(Enum):
    Current = ""
    TransparentBackground = "TransparentBackground"
    BlackBackground = "BlackBackground"
    DefaultBackground = "DefaultBackground"
    GradientBackground = "GradientBackground"
    GrayBackground = "GrayBackground"
    PrintBackground = "PrintBackground"
    WhiteBackground = "WhiteBackground"


class Camera(object):
    def __init__(self, render_view):
        self._render_view = render_view
        self._camera = render_view.GetActiveCamera()

    def _validate_vec3(self, pos, name):
        if not isinstance(pos, (tuple, list)):
            raise TypeError('%s should be tuple or list' % name)

        if len(pos) != 3:
            raise ValueError('%s must have 3 components' % name)

    def dolly(self, value):
        """
        Divide the camera's distance from the focal point by the given dolly
        value
        """
        self._camera.Dolly(value)
        Render(self._render_view)

    @property
    def roll(self):
        """
        Rotate the camera about the direction of projection.
        """
        return self._camera.GetRoll()

    @roll.setter
    def roll(self, angle):
        self._camera.SetRoll(angle)
        Render(self._render_view)

    def azimuth(self, angle):
        """
        Rotate the camera about the view up vector centered at the focal point.
        """
        self._camera.Azimuth(angle)
        Render(self._render_view)

    def yaw(self, angle):
        """
        Rotate the focal point about the view up vector, using the camera's
        position as the center of rotation.
        """
        self._camera.Yaw(angle)
        Render(self._render_view)

    def elevation(self, angle):
        """
        Rotate the camera about the cross product of the negative of the
        direction of projection and the view up vector, using the focal point as
        the center of rotation.
        """
        self._camera.Elevation(angle)
        Render(self._render_view)

    def pitch(self, angle):
        """
        Rotate the focal point about the cross product of the view up vector and
        the direction of projection, using the camera's position as the center
        of rotation.
        """
        self._camera.Pitch(angle)
        Render(self._render_view)

    @property
    def position(self):
        """
        Set/Get the position of the camera in world coordinates.
        """
        return self._camera.GetPosition()

    @position.setter
    def position(self, pos):
        self._validate_vec3(pos, 'Position')
        [x, y, z] = pos
        self._camera.SetPosition(x, y, z)
        Render(self._render_view)

    @property
    def focal_point(self):
        """
        Set/Get the focal of the camera in world coordinates.
        """
        return self._camera.GetFocalPoint()

    @focal_point.setter
    def focal_point(self, point):
        self._validate_vec3(point, 'Point')
        [x, y, z] = point
        self._camera.SetFocalPoint(x, y, z)
        Render(self._render_view)

    @property
    def view_up(self):
        """
        Set/Get the view up direction for the camera.
        """
        return self._camera.GetViewUp()

    @view_up.setter
    def view_up(self, direction):
        self._validate_vec3(direction, 'Direction')
        [vx, vy, vz] = direction
        self._camera.SetViewUp(vx, vy, vz)
        Render(self._render_view)

    @property
    def distance(self):
        """
        Move the focal point so that it is the specified distance from the
        camera position
        """
        return self._camera.GetDistance()

    @distance.setter
    def distance(self, distance):
        self._camera.SetDistance(distance)
        Render(self._render_view)

    def zoom(self, factor):
        """
        Decrease the view angle by the specified factor. A value greater than 1
        is a zoom-in, a value less than 1 is a zoom-out.
        """
        self._camera.Zoom(factor)
        Render(self._render_view)


class View(object):
    def __init__(self, render_view):
        self._render_view = render_view

    def save_screenshot(self, file_path, palette=Palette.Current, width=-1,
                        height=-1, resolution=None):
        options = {
            'OverrideColorPalette': palette.value
        }

        if width > 0 and height > 0:
            options['ImageResolution'] = (width, height)

        if resolution is not None:
            if not isinstance(resolution, (tuple, list)):
                raise TypeError('Resolution must be a tuple or list')
            if len(resolution) != 2:
                raise ValueError('Resolution must have 2 components')
            options['ImageResolution'] = resolution

        SaveScreenshot(file_path, viewOrLayout=self._render_view, **options)

    @property
    def camera(self):
        return Camera(self._render_view)



from tomviz import utils


class Dataset:
    def __init__(self, data_object, data_source):
        self._data_object = data_object
        self._data_source = data_source

        self._tilt_axis = 2

    @property
    def active_scalars(self):
        return utils.get_array(self._data_object)

    @active_scalars.setter
    def active_scalars(self, v):
        utils.set_array(self._data_object, v)

    @property
    def scalars_names(self):
        return utils.array_names(self._data_object)

    def scalars(self, name=None):
        return utils.get_array(self._data_object, name)

    @property
    def spacing(self):
        return utils.get_spacing(self._data_object)

    @spacing.setter
    def spacing(self, v):
        utils.set_spacing(self._data_object, v[0], v[1], v[2])

    @property
    def tilt_angles(self):
        return utils.get_tilt_angles(self._data_object)

    @tilt_angles.setter
    def tilt_angles(self, v):
        utils.set_tilt_angles(self._data_object, v)

    @property
    def tilt_axis(self):
        return self._tilt_axis

    @tilt_axis.setter
    def tilt_axis(self, v):
        self._tilt_axis = v

    @property
    def dark(self):
        if not self._data_source.dark_data:
            return None
        return utils.get_array(self._data_source.dark_data)

    @property
    def white(self):
        if not self._data_source.white_data:
            return None
        return utils.get_array(self._data_source.white_data)

    @property
    def file_name(self):
        return self._data_source.file_name

    @property
    def metadata(self):
        return self._data_source.metadata

    def create_child_dataset(self):
        new_data = utils.make_child_dataset(self._data_object)
        return Dataset(new_data, self._data_source)


def create_dataset(data_object, data_source):
    return Dataset(data_object, data_source)



r"""
Currently, this package is experimental and may change in the future.
"""
from __future__ import absolute_import


#------------------------------------------------------------------------------
# this little trick is for static builds of VTK. In such builds, if
# the user imports this Python package in a non-statically linked Python
# interpreter i.e. not of the of the VTK-python executables, then we import the
# static components importer module.
try:
    from . import vtkCommonCore
except ImportError:
    import _vtkmodules_static
#------------------------------------------------------------------------------



import http.server
import socketserver
from urllib.parse import urlparse
from urllib.parse import parse_qs
from os import chdir
from os import listdir
from os import path
from os.path import relpath
from os import getcwd
from os import access
from os import R_OK
import pathlib
import json
import cinemasci

#
# global variables - can't seem to add an instance variable to the
# subclass of SimpleHTTPRequestHandler
#
CinemaInstallPath = "CINEMAJUNK"
ServerInstallPath = "CINEMAJUNK"

def set_install_path():
    global CinemaInstallPath
    global ServerInstallPath

    ServerInstallPath = str(pathlib.Path(__file__).parent.absolute())
    CinemaInstallPath = str(pathlib.Path(__file__).parent.absolute())
    # edit the path to get the correct installation path
    CinemaInstallPath = CinemaInstallPath.strip("/server")
    CinemaInstallPath = "/" + CinemaInstallPath + "/viewers"

def verify_cinema_database( viewer, cdb, assetname ):
    result = False

    if viewer == "view":
        # this is the default case
        if assetname is None:
            assetname = "FILE"

        db = cinemasci.new("cdb", {"path": cdb})
        db.read_data_from_file()

        if db.parameter_exists(assetname):
            result = True
        else:
            print("")
            print("ERROR: Cinema viewer \'view\' is looking for a column named \'{}\', but the".format(assetname)) 
            print("       the cinema database \'{}\' doesn't have one.".format(cdb)) 
            print("")
            print("       use \'--assetname <name>\' where <name> is one of these possible values") 
            print("       that were found in \'{}\':".format(cdb))
            print("")
            print("           \"" + ' '.join(db.get_parameter_names()) + "\"")
            print("")
    else:
        result = True

    return result 

#
# CinemaSimpleReqestHandler
#
# Processes GET requests to find viewers and databases
#
class CinemaSimpleRequestHandler(http.server.SimpleHTTPRequestHandler):

    @property
    def verbose(self):
        return self._verbose

    @verbose.setter
    def verbose(self, value):
        self._verbose = value

    @property
    def assetname(self):
        return self._assetname

    @assetname.setter
    def assetname(self, value):
        self._assetname = value

    @property
    def viewer(self):
        return self._viewer

    @viewer.setter
    def viewer(self, value):
        self._viewer = value

    def translate_path(self, path ):
        # print("translate_path: {}".format(path))
        return path

    def log(self, message):
        if self.verbose:
            print(message)

    def log_message( self, format, *args ):
        pass

    def do_GET(self):
        global CinemaInstallPath
        global ServerInstallPath

        self.log(" ")
        self.log("PATH     : {}".format(self.path))

        # the request is for a viewer
        if self.path == "/":
            if not path.isdir(self.base_path):
                self.log("ERROR")
                self.path = ServerInstallPath + "/error_no-database.html"
                return http.server.SimpleHTTPRequestHandler.do_GET(self)

            elif self.viewer == "explorer": 
                # handle a request for the Cinema:Explorer viewer
                self.log("EXPLORER")
                self.path = CinemaInstallPath + "/cinema_explorer.html"
                self.log(self.path)
                return http.server.SimpleHTTPRequestHandler.do_GET(self)

            elif self.viewer == "view": 
                # handle a request for the Cinema:View viewer
                self.log("VIEW")
                self.path = CinemaInstallPath + "/cinema_view.html"
                return http.server.SimpleHTTPRequestHandler.do_GET(self)

            if self.viewer == "test": 
                # handle a request for the Cinema:Explorer viewer
                self.log("TEST")
                self.path = ServerInstallPath + "/cinema_test.html"
                return http.server.SimpleHTTPRequestHandler.do_GET(self)

            else:
                self.log("VIEWER: {}".format(self.viewer))


        if self.path.endswith("cinema_attributes.json"):
            # this is a request to the server for attributes
            self.log("ATTRIBUTE REQUEST") 

            if (not self.assetname == None):
                json = "{{\"assetname\" : \"{}\"}}".format(self.assetname)
                self.send_response(200)
                self.send_header("Content-type", "text/html")
                self.send_header("Content-length", len(json))
                self.end_headers()
                self.wfile.write(str.encode(json))
            return

        if self.path.endswith("databases.json"):
            self.log("DATABASES   : {}".format(self.path))
            json = self.create_database_list()

            self.send_response(200)
            self.send_header("Content-type", "text/html")
            self.send_header("Content-length", len(json))
            self.end_headers()
            self.wfile.write(str.encode(json))
            return


        if self.path.startswith("/cinema/"):
            # handle a requests for sub components of the viewers 
            # NOTE: fragile - requires 'cinema' path be unique

            self.log("CINEMA   : {}".format(self.path))
            self.path = CinemaInstallPath + self.path 
            self.log("         : {}".format(self.path))
            return http.server.SimpleHTTPRequestHandler.do_GET(self)

        else:
            # everything else
            self.log("NORMAL   : {}".format(self.path))
            self.path = relpath(self.path.strip("/"), getcwd())
            # self.log("UPDATED  : {}".format(self.path))
            return http.server.SimpleHTTPRequestHandler.do_GET(self)

    def create_database_list( self ):
        json_string = ""
        if self.viewer == "view":
            dbs = [self.base_path]
            cdbname = path.basename(dbs[0])
            json_string = "[{{ \"database_name\": \"{}\", \"datasets\": [ ".format(cdbname)

            for db in dbs:
                cdbname = path.basename(db)
                json_string += "{{ \"name\" : \"{}\", \"location\" : \"{}\" }},".format(cdbname, db)

            # remove the last comma
            json_string = json_string[:-1]
            # close the string
            json_string += "]}]"

        elif self.viewer == "explorer":
            dbs = [self.base_path]
            cdbname = path.basename(dbs[0])
            json_string = "["

            for db in dbs:
                cdbname = path.basename(db)
                json_string += "{{ \"name\" : \"{}\", \"directory\" : \"{}\" }},".format(cdbname, db)

            # remove the last comma
            json_string = json_string[:-1]
            # close the string
            json_string += "]"

        else:
            self.log("ERROR: invalid view type {}".format(this.viewer))

        return json_string 

def run_cinema_server( viewer, data, port, assetname="FILE"):
    localhost = "http://127.0.0.1"

    fullpath  = path.abspath(data)
    datadir   = path.dirname(fullpath)
    cinemadir = path.basename(fullpath)

    chdir(datadir)

    if verify_cinema_database(viewer, cinemadir, assetname) :
        set_install_path()
        cin_handler = CinemaSimpleRequestHandler
        cin_handler.base_path = cinemadir
        cin_handler.verbose   = False
        cin_handler.viewer    = viewer
        cin_handler.assetname = assetname
        with socketserver.TCPServer(("", port), cin_handler) as httpd:
            urlstring = "{}:{}".format(localhost, port)
            print(urlstring)
            httpd.serve_forever()




def transform(dataset, axis, ratio, value):
    try:
        import tomopy
    except ImportError:
        print('Failed to import tomopy, which is required for this operator')
        raise

    array = dataset.active_scalars

    result = tomopy.misc.corr.circ_mask(array, axis, ratio, value)

    dataset.active_scalars = result



import os
import numpy as np
import tomviz.utils

from vtk import vtkImageData

_installed = False
DEFAULT_URL = "https://tiled.nsls2.bnl.gov"
TILED_URL = os.getenv("TILED_URL", DEFAULT_URL)
try:
    from tiled.client import from_uri
    c = from_uri(TILED_URL, "dask")
    from databroker.queries import TimeRange
    _installed = True
except ImportError:
    pass


def installed():
    return _installed


def catalogs():
    cats = []
    for name in c:
        cats.append({
            "name": name,
            "description": "blank" #cat.describe()['description']
        })

    cats = sorted(cats, key=lambda c: c['name'])

    return cats


def runs(catalog_name, since, until, limit):
    runs = []

    current = c[catalog_name]

    if since != "" and until != "":
        current = current.search(TimeRange(since=since, until=until))

    cat = current.values_indexer[-1:-limit:-1]

    for run in cat:
        md = run.metadata
        runs.append({
            "uid": md['start']['uid'],
            "planName": md['start']['plan_name'],
            "scanId": md['start']['scan_id'],
            "startTime": md['start']['time'],
        })

    runs = sorted(runs, key=lambda r: r['uid'])

    return runs


def tables(catalog_name, run_uid):
    tables = []
    for name, _ in c[catalog_name][run_uid].items():
        tables.append({
            "name": name,
        })

    tables = sorted(tables, key=lambda t: t['name'])

    return tables


def variables(catalog_name, run_uid, table):
    variables = []

    # Would be nice to use context manager here, but it doesn't seem to close
    # the dataset.
    dataset = c[catalog_name][run_uid][table].read()
    for name, variable in dataset.data_vars.items():
        variables.append({
            "name": name,
            "size": variable.data.nbytes
        })

    variables = sorted(variables, key=lambda v: v['name'])

    return variables


def load_variable(catalog_name, run_uid, table, variable):
    if run_uid not in c[catalog_name]:
        raise Exception(f"Unable to load run: {run_uid}")

    if table not in c[catalog_name][run_uid]:
        raise Exception(f"Unable to find table: {table}")

    if variable not in c[catalog_name][run_uid][table]['data']:
        raise Exception(f"Unable to find variable: {variable}")

    run = c[catalog_name][run_uid]
    data = run[table]['data'][variable].data
    shape = data.shape
    data = data.reshape((shape[0]*shape[1], shape[2], shape[3]))
    # Convert to numpy array
    data = np.ascontiguousarray(data)

    image_data = vtkImageData()
    (x, y, z) = data.shape

    # Metadata?
    image_data.SetOrigin(0, 0, 0)
    image_data.SetSpacing(1, 1, 1)
    image_data.SetExtent(0, x - 1, 0, y - 1, 0, z - 1)
    tomviz.utils.set_array(image_data, data)

    # Now load the angles
    if 'zps_pi_r_monitor' not in run or \
       'data' not in run['zps_pi_r_monitor'] or \
       'zps_pi_r' not in run['zps_pi_r_monitor']['data']:
        raise Exception("No angles found!")

    angles = run['zps_pi_r_monitor']['data']['zps_pi_r'].data
    tomviz.utils.set_tilt_angles(image_data, angles)

    return image_data



import tomviz.operators

NUMBER_OF_CHUNKS = 10


class SquareRootOperator(tomviz.operators.CancelableOperator):

    def transform(self, dataset):
        """Define this method for Python operators that
        transform input scalars"""

        import numpy as np
        self.progress.maximum = NUMBER_OF_CHUNKS

        scalars = dataset.active_scalars
        if scalars is None:
            raise RuntimeError("No scalars found!")

        if scalars.min() < 0:
            print("WARNING: Square root of negative values results in NaN!")
        else:
            # transform the dataset
            # Process dataset in chunks so the user gets an opportunity to
            # cancel.
            result = np.float32(scalars)
            step = 0
            for chunk in np.array_split(result, NUMBER_OF_CHUNKS):
                if self.canceled:
                    return
                np.sqrt(chunk, chunk)
                step += 1
                self.progress.value = step

            # set the result as the new scalars.
            dataset.active_scalars = result



from pyxrf.api_dev import make_hdf

from xrf_tomo import create_log_file


def make_hdf5(start_scan, stop_scan, working_directory, successful_scans_only,
              log_file_name):

    kwargs = {
        'start': start_scan,
        'end': stop_scan,
        'wd': working_directory,
        'successful_scans_only': successful_scans_only,
    }
    make_hdf(**kwargs)

    kwargs = {
        'fn_log': log_file_name,
        'wd': working_directory,
    }
    create_log_file(**kwargs)



import tomviz.operators


class PeronaMalikAnisotropicDiffusion(tomviz.operators.CancelableOperator):

    def transform(self, dataset, conductance=1.0, iterations=100,
                  timestep=0.0625):
        """This filter performs anisotropic diffusion on an image using
        the classic Perona-Malik, gradient magnitude-based equation.
        """

        # Initial progress
        self.progress.value = 0
        self.progress.maximum = 100

        # Approximate percentage of work completed after each step in the
        # transform
        STEP_PCT = [10, 20, 90, 100]

        try:
            import numpy as np
            import itk
            from tomviz import itkutils
        except Exception as exc:
            print("Could not import necessary module(s)")
            raise exc

        try:
            self.progress.value = STEP_PCT[0]
            self.progress.message = "Converting data to ITK image"

            # Get the ITK image. The itk.GradientAnisotropicDiffusionImageFilter
            # is templated over float pixel types only, so explicitly request a
            # float ITK image type.
            array = dataset.active_scalars.astype(np.float32)
            itk_image = itk.GetImageViewFromArray(array)
            itk_image.SetSpacing(dataset.spacing)
            itk_image_type = type(itk_image)

            self.progress.value = STEP_PCT[1]
            self.progress.message = "Running filter"

            DiffusionFilterType = \
                itk.GradientAnisotropicDiffusionImageFilter[itk_image_type,
                                                            itk_image_type]
            diffusion_filter = DiffusionFilterType.New()
            diffusion_filter.SetConductanceParameter(conductance)
            diffusion_filter.SetNumberOfIterations(iterations)
            diffusion_filter.SetTimeStep(timestep)
            diffusion_filter.SetInput(itk_image)
            itkutils.observe_filter_progress(self, diffusion_filter,
                                             STEP_PCT[1], STEP_PCT[2])

            try:
                diffusion_filter.Update()
            except RuntimeError:
                return

            self.progress.message = "Saving results"

            itkutils.set_itk_image_on_dataset(diffusion_filter.GetOutput(),
                                              dataset)

            self.progress.value = STEP_PCT[3]
        except Exception as exc:
            print("Problem encountered while running %s" %
                  self.__class__.__name__)
            raise exc



import copy
import logging

from tomviz import executor

from .load_data_source import load_data_source

logger = logging.getLogger('tomviz')


def data_source_dependencies(data_source):
    deps = []
    for op in data_source.get('operators', []):
        for value in op.get('arguments', {}).values():
            if isinstance(value, str) and value.startswith('0x'):
                deps.append(value)

        for child in op.get('dataSources', []):
            deps += data_source_dependencies(child)

    return deps


class DependencyInfo:
    def __init__(self, dep, data_sources):
        self.dep = dep
        self.data_sources = data_sources
        self._path = None

    @property
    def path(self):
        # After it has been computed once, cache the path.
        if self._path is not None:
            return self._path

        path = []

        def recurse(data_sources):
            for ds in data_sources:
                if ds.get('id', '') == self.dep:
                    path.insert(0, ds)
                    return True

                for op in ds.get('operators', []):
                    children = op.get('dataSources', [])
                    if recurse(children):
                        path.insert(0, ds)
                        return True

        recurse(self.data_sources)
        self._path = path
        return path

    @property
    def root_data_source(self):
        return self.path[0]

    @property
    def depth(self):
        return len(self.path) - 1

    @property
    def is_root(self):
        return self.depth == 0

    @property
    def is_child(self):
        return self.depth > 0


def load_dependencies(target_source, state, state_file_path, progress_method,
                      progress_path):
    already_loaded = {}
    data_sources = state.get('dataSources', [])

    def local_load_data_source(data_source):
        id = data_source.get('id', '')
        if id not in already_loaded:
            already_loaded[id] = load_data_source(data_source, state_file_path)

        return already_loaded[id]

    def local_load_dependencies(data_source):
        deps = data_source_dependencies(data_source)
        if not deps:
            # Nothing to do
            return already_loaded

        for dep in deps:
            # Load the root data source of this dependency
            info = DependencyInfo(dep, data_sources)
            root_source = info.root_data_source

            # First, load in the root source
            root_dataset = local_load_data_source(root_source)

            # Load any dependencies of this data source
            local_load_dependencies(root_source)

            # If it is a child dependency, execute the pipeline
            if info.is_child:
                # Make sure it has a depth of 1. Otherwise, raise an
                # exception, as we only support a depth of 1.
                if info.depth != 1:
                    msg = (
                        f'Dependency {dep} has a depth of {info.depth}, '
                        'but only a depth of 0 or 1 is currently supported '
                    )
                    raise Exception(msg)

                # Execute the pipeline. The result is what we need.
                with executor._progress(progress_method,
                                        progress_path) as progress:
                    progress.started()
                    logger.info(
                        f'Running pipeline to create child dependency: {dep}')

                    # The operator might modify the dataset in place.
                    # We also might still need the root dataset as it is, so
                    # make a deep copy.
                    data = copy.deepcopy(root_dataset)
                    kwargs = {
                        'data': data,
                        'operators': root_source['operators'],
                        'start_at': 0,
                        'progress': progress,
                        'child_output_path': None,
                        'dataset_dependencies': already_loaded,
                    }

                    result = executor.run_pipeline(**kwargs)
                    if result is None:
                        # The data was modified in place
                        result = data

                    already_loaded[dep] = result

                    logger.info(f'Child dependency created: {dep}')
                    progress.finished()

        return already_loaded

    return local_load_dependencies(target_source)
